"""The Replit Python module."""

from .audio import Audio
from .database import db, Database
from .users import get_user, User

# Backwards compatibility.
from .termutils import clear


audio = Audio()



"""Python object representations of HTML."""

from abc import ABC

import flask


class HTMLElement(ABC):
    """Base class for an HTML element."""

    def __init__(self, content: str, **attrs: str) -> None:
        """Initialize the class.

        Args:
            content (str): The content inside the HTML tag.
            attrs (str): The attributes of tag.
        """
        self.content = content
        self.attrs = attrs

    def stringify_attrs(self) -> str:
        """Generate a string version of the tags attributes.

        Returns:
            str: A string in the form of attr1="value1" attr2="value2"
        """
        return " ".join(f'{name}="{value}"' for name, value in self.attrs.items())

    def __str__(self) -> str:
        """Generates an HTML representation of the tag."""
        attr_text = self.stringify_attrs()
        return (
            f"<{self.tagname}{' ' + attr_text if attr_text else ''}>{self.content}"
            f"</{self.tagname}>"
        )


class Paragraph(HTMLElement):
    """Represents a Paragraph (p) tag."""

    tagname = "p"


class Link(HTMLElement):
    """Represents a Link (a) tag."""

    tagname = "a"

    def __init__(self, content: str, href: str, **attrs: str) -> None:
        """Initialize the class.

        Args:
            content (str): The content of the link tag.
            href (str): The href attribute, where the link points to.
            attrs (str): Any additional attributes of the tag.
        """
        super().__init__(content, **{"href": href, **attrs})


class Page(flask.Response):
    """Represents an HTML page."""

    def __init__(self, title: str = None, head: str = "", body: str = "") -> None:
        """Initialize the class.

        Args:
            title (str): The title of the page. If not provided no title tag will be
                added.
            head (str): The HTML to put in the head of the page. Defaults to nothing.
            body (str): The HTML to put in the body of the page. Defaults to nothing.
        """
        self.title = title
        self.head = head
        self.body = body

        title_html = f"<title>{self.title}</title>\n    " if self.title else ""
        super().__init__(
            f"""<!DOCTYPE html>
            <html>
            <head>
                {title_html}{self.head}
            </head>
            <body>
                {self.body}
            </body>
            </html>"""
        )



"""Interface with the Replit Database."""
import os
from sys import stderr
from typing import Optional

from .database import AsyncDatabase, Database
from .jsonkey import AsyncJSONKey, JSONKey

__all__ = ["AsyncDatabase", "Database", "AsyncJSONKey", "JSONKey"]


db: Optional[Database]
db_url = os.environ.get("REPLIT_DB_URL")
if db_url:
    db = Database(db_url)
else:
    print(
        "Warning: REPLIT_DB_URL does not exist, are we running on repl.it? "
        "Database will not function.",
        file=stderr,
    )
    db = None



"""Core of flask_tools."""

from dataclasses import dataclass
from functools import wraps
from pathlib import Path
from typing import Any, Callable, List, Set

import flask

from .utils import sign_in


@dataclass
class ReplitUserContext:
    """A dataclass defining a Replit Auth state."""

    user_id: int
    name: str
    roles: str

    @classmethod
    def from_headers(cls, headers: dict) -> Any:
        """Initialize an instance using the Replit identification headers.

        Args:
            headers (dict): A dictionary of headers received

        Returns:
            Any: An initialized class instance
        """
        return cls(
            user_id=headers.get("X-Replit-User-Id"),
            name=headers.get("X-Replit-User-Name"),
            roles=headers.get("X-Replit-User-Roles"),
        )

    @property
    def is_authenticated(self) -> bool:
        """Check whether the user is authenticated in with Replit Auth.

        Returns:
            bool: whether or not the authentication is activated.
        """
        return bool(self.name)


class ReplitRequest(flask.Request):
    """Represents a client request."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initializes request and runs update_auth.

        Args:
            args (Any): The arguments to be passed to the superclass.
            kwargs (Any): The keyword arguments to be passed to the superclass.
        """
        super().__init__(*args, **kwargs)
        self.update_auth()

    def update_auth(self) -> None:
        """Update the user_info property to be a ReplitUserContext."""
        self.user_info = ReplitUserContext.from_headers(self.headers)

    @property
    def is_authenticated(self) -> bool:
        """Check whether the user is authenticated with Replit.

        Returns:
            bool: Whether or not the user is signed in
        """
        return self.user_info.is_authenticated


class ReplitApp(flask.Flask):
    """Represents a web application."""

    request_class = ReplitRequest

    def __init__(
        self, import_name: str, nice_jinja: bool = True, **kwargs: Any
    ) -> None:
        """Initialize the app.

        Args:
            import_name (str): The name of the app, usually __name__
            nice_jinja (bool): Whether to change jinja settings to make them
                prettier. Defaults to True.
            **kwargs (Any): Extra keyword arguments to be passed to the flask init
                function.
        """
        super().__init__(import_name, **kwargs)
        if nice_jinja:
            self.jinja_env.trim_blocks = True
            self.jinja_env.lstrip_blocks = True

    def login_wall(
        self,
        exclude: Set[str] = ("/",),
        handler: Callable = None,
    ) -> None:
        """Require users to be logged-in on all pages.

        Args:
            exclude (Tuple[str]): The routes that should not require sign in.
                Defaults to just /.
            handler (Callable): The handler to call when the user is not signed in. If
                not provided, defaults to maqpy.signin()
        """
        self._lw_exclude = set(exclude) or set()
        self._lw_handler = handler or (lambda: sign_in())

    def _request_handler(self, rule: str, view_func: Callable) -> Callable:
        """Return a handler for a given request.

        This enables the all_pages_sign_in feature.

        Args:
            rule (str): The url that the route will be matched to
            view_func (Callable): The original view function that will be called.

        Returns:
            Callable: A handler that runs the middleware and calls the original function
        """

        @wraps(view_func)
        def handler(*args: Any, **kwargs: Any) -> Any:
            if (
                hasattr(self, "_lw_exclude")
                and self._lw_exclude is not None
                and rule not in self._lw_exclude
                and not flask.request.signed_in
            ):
                return self._lw_handler(*args, **kwargs)
            return view_func(*args, **kwargs)

        return handler

    def add_url_rule(
        self,
        rule: str,
        endpoint: str = None,
        view_func: Callable = None,
        provide_automatic_options: bool = None,
        **options: Any
    ) -> None:
        """Replaces view function with custom handler."""
        return super().add_url_rule(
            rule,
            endpoint=endpoint,
            view_func=self._request_handler(rule, view_func)
            if view_func is not None
            else view_func,
            provide_automatic_options=provide_automatic_options,
            **options
        )

    def _run(self, *args: Any, **kwargs: Any) -> Any:
        """Interface with the underlying Flask instance's run function.

        Args:
            args (Any): The arguments to be passed to the superclass' run method.
            kwargs (Any): The keyword arguments to be passed to the superclass' run
                method.

        Returns:
            Any: The result of running the superclasses' run method.
        """
        return super().run(*args, **kwargs)

    def run(self, port: int = 8080, localhost: bool = False, **kwargs: Any) -> None:
        """Run the app.

        Args:
            port (int): The port to run the app on. Defaults to 8080.
            localhost (bool): Whether to run the app without exposing it on all
                interfaces. Defaults to False.
            **kwargs (Any): Extra keyword arguments to be passed to the flask app's run
                method.
        """
        super().run(host="localhost" if localhost else "0.0.0.0", port=port, **kwargs)

    def debug(
        self,
        watch_dirs: List[str] = None,
        watch_files: List[str] = None,
        port: int = 8080,
        localhost: bool = False,
        **kwargs: Any
    ) -> None:
        """Run the app in debug mode.

        Args:
            watch_dirs (List[str]): Directories whose files will be added to
                watch_files. Defaults to [].
            watch_files (List[str]): Files to watch, and if changes are detected
                the server will be restarted. Defaults to [].
            port (int): The port to run the app on. Defaults to 8080.
            localhost (bool): Whether to run the app without exposing it on all
                interfaces. Defaults to False.
            **kwargs (Any): Extra keyword arguments to be passed to the flask app's run
                method.
        """
        watch_files = list(watch_files or [])

        for directory in watch_dirs or []:
            if not isinstance(directory, Path):
                directory = Path(directory)
            watch_files += [str(f) for f in directory.iterdir() if f.is_file()]

        super().run(
            host="localhost" if localhost else "0.0.0.0",
            port=port,
            debug=True,
            extra_files=watch_files,
        )



"""Utitilities to make development easier."""

from functools import wraps
import time
from typing import Any, Callable, Iterable, Optional, Union

import flask
from werkzeug.local import LocalProxy

from .html import Page


sign_in_snippet = (
    '<script authed="location.reload()" '
    'src="https://auth.turbio.repl.co/script.js"></script>'
)


def sign_in(title: str = "Please Sign In") -> Page:
    """Return a sign-in page.

    Args:
        title (str): The title of the sign in page. Defaults to "Please Sign In".

    Returns:
        Page: The sign-in page.
    """
    return Page(title=title, body=sign_in_snippet)


sign_in_page = sign_in()


def needs_sign_in(func: Callable = None, login_res: str = sign_in_page) -> Callable:
    """A decorator that enforces that the user is signed in before accessing the page.

    Args:
        func (Callable): The function passed in if used as a decorator. Defaults to
            None.
        login_res (str): The HTML to show when the user needs to sign in. Defaults to
            sign_in_snippet.

    Returns:
        Callable: The new handler.
    """

    def decorator(func: Callable) -> Callable:
        @wraps(func)
        def handler(*args: Any, **kwargs: Any) -> flask.Response:
            if flask.request.signed_in:
                return func(*args, **kwargs)
            else:
                return login_res

        return handler

    if func is not None:  # called with no options @needs_signin
        return decorator(func)
    else:  # called with options, eg @needs_signin(login_html='...')
        return decorator


def needs_params(
    *param_names: str,
    src: Union[str, dict] = "form",
    onerror: Callable[[str], flask.Response] = None,
) -> Callable:
    """Require paramaters before a handler can be activated.

    Args:
        param_names (str): The paramaters that must be in the request.
        src (Union[str, dict]): The source to get the paramaters from. Can be "form"
            to use flask.request.form (POST requests), "query" for flask.request.query
            (GET requests), or a custom dictionary.
        onerror (Callable): A function to handle when a paramater is missing. It will
            be passed the parameter that is missing. If no function is specified a
            handler that returns a descriptive error and 400 Bad Request status code
            will be used.

    Raises:
        TypeError: No paramaters were provided or an invalid one was provided.

    Returns:
        Callable: The new handler.
    """
    if len(param_names) < 1:
        raise TypeError("You must specify at least one required paramater name")
    # If function is used as a decorator with no arguments, the first argument will be
    # a function, so type check all of the param names to catch mistakes
    if not all(isinstance(p, str) for p in param_names):
        raise TypeError("All paramater names should be strings.")

    def default_onerror(missing_param: str) -> flask.Response:
        return flask.Response(
            f"Parameter {missing_param!r} is required but is missing",
            400,
            mimetype="text/plain",
        )

    onerror = default_onerror if onerror is None else onerror

    def decorator(func: Callable) -> Callable:
        @wraps(func)
        def handler(*args: Any, **ignoredkwargs: Any) -> flask.Response:
            if src == "form":
                params = flask.request.form
            elif src == "query":
                params = flask.request.args
            else:
                params = src
            param_kwargs = {}
            for p in param_names:
                if p not in params:
                    return onerror(p)
                param_kwargs[p] = params[p]
            return func(*args, **param_kwargs)

        return handler

    return decorator


def local_redirect(location: str, code: int = 302) -> flask.Response:
    """Perform a redirection to a local path without downgrading to HTTP.

    Args:
        location (str): The path to redirect to.
        code (int): The code to use for the redirect. Defaults to 302.

    Returns:
        flask.Response: The redirect response.
    """
    # Use a LocalProxy so that it can be called before the request context is available
    return LocalProxy(
        lambda: flask.redirect(
            "https://" + flask.request.headers["host"] + location, code
        )
    )


def authed_ratelimit(
    max_requests: int,
    period: float,
    login_res: str = sign_in_page,
    get_ratelimited_res: Callable[[float], str] = (
        lambda left: f"Too many requests, wait {left} sec"
    ),
) -> Callable[[Callable], flask.Response]:
    """Require sign in and limit the amount of requests each signed in user can perform.

    This decorator also calls needs_signin for you and passes the login_res kwarg
        directly to it.

    Args:
        max_requests (int): The maximum amount of requests allowed in the period.
        period (float): The length of the period.
        login_res (str): The response to be shown if the user is not signed in, passed
            to needs_sign_in.
        get_ratelimited_res (Callable[[float], str]): A callable which is passed the
            amount of time remaining before the user can request again and returns the
            response that should be sent to the user.

    Returns:
        Callable[[Callable], flask.Response]: A function which decorates the handler.
    """

    def decorator(func: Callable) -> flask.Response:
        last_reset = time.time()
        num_requests = {}

        # Checks for signin first, before checking ratelimit
        @needs_sign_in(login_res=login_res)
        @wraps(func)
        def handler(*args: Any, **kwargs: Any) -> flask.Response:
            nonlocal last_reset
            nonlocal num_requests

            name = flask.request.auth.name
            now = time.time()

            if now - last_reset >= period:
                last_reset = now
                num_requests = {}

            times_requested = num_requests.get(name, 0)
            if times_requested >= max_requests:
                res = get_ratelimited_res(period - (now - last_reset))
                # Make a reponse object so that status can be set
                if not isinstance(res, flask.Response):
                    res = flask.make_response(res)
                res.status = "429"
                return res

            num_requests[name] = times_requested + 1
            return func(*args, **kwargs)

        return handler

    return decorator


def find(
    data: Iterable, cond: Callable[[Any], bool], allow_multiple: bool = False
) -> Optional[Any]:
    """Find an item in an iterable.

    Args:
        data (Iterable): The iterable to search through.
        cond (Callable[[Any], bool]): The function to call for each item to check if it
            is a match.
        allow_multiple (bool): If multiple result are found, return the first one if
            allow_multiple is True, otherwise return None.

    Returns:
        Optional[Any]: The item if exactly one match was found, otherwise None.
    """
    matches = [item for item in data if cond(item)]
    if len(matches) > 1:
        return matches[0] if allow_multiple else None
    return matches[0] if len(matches) == 1 else None


def chain_decorators(*decorators: Callable[[Callable], Any]) -> Callable:
    """Return a decorator that applies each of the decorators to the function.

    Args:
        *decorators (Callable[[Callable], Any]): The decorators to apply to the
            function. They are treated as if they are written in the order they appear.

    Raises:
        TypeError: If no decorators are passed.

    Returns:
        Callable: A decorator function.
    """

    def dec(func: Callable) -> Callable:
        for decorator in reversed(list(decorators) + [wraps(func)]):
            func = decorator(func)
        return func

    if not decorators:
        raise TypeError("You must provide at least one decorator to chain")
    return dec



# flake8: noqa
from typing import List
from typing_extensions import TypedDict
from enum import Enum


class ReaderType(Enum):
    "An Enum for the types of sources."

    def __str__(self) -> str:
        return self._value_

    def __repr__(self) -> str:
        return f"ReaderType.{self._name_}"

    wav_file = "wav"
    "ReaderType : The type for a .wav file."
    aiff_file = "aiff"
    "ReaderType : The type for a .aiff file."
    mp3_file = "mp3"
    "ReaderType : The type for a .mp3 file."
    tone = "tone"
    "ReaderType : The type for a generated tone."


class WaveType(Enum):
    "The different wave shapes that can be used for tone generation."

    def __str__(self) -> str:
        return self._value_

    WaveSine = 0
    "WaveType : The WaveSine wave shape."
    WaveTriangle = 1
    "WaveType : The Triangle wave shape."
    WaveSaw = 2
    "WaveType : The Saw wave shape."
    WaveSqr = 3
    "WaveType : The Square wave shape."


file_types: List[ReaderType] = [
    ReaderType.aiff_file,
    ReaderType.wav_file,
    ReaderType.mp3_file,
]
"The different file types for sources in a list."


class RequestArgs(TypedDict, total=False):
    "The additional arguments for a request that are type-specific."
    Pitch: float
    "float : The pitch/frequency of the tone. Only used if the request type is tone."
    Seconds: float

    "float : The duration for the tone to be played. Only used if the request type is tone."
    WaveType: WaveType or int
    "WaveType : The wave type of the tone. Only used if the request type is tone."
    Path: str
    "str : The path to the file to be read. Only used if the request is for a file type."


class RequestData(TypedDict):
    "A request to pid1 for a source to be played."
    ID: int
    "int : The ID of the source. Only used for updating a pre-existing source."
    Paused: bool or None
    "bool or None : Wether the source with the provided ID should be paused or not.  Can only be used when updating a source."
    Volume: float
    "float : The volume the source should be played at. (1 being 100%)"
    DoesLoop: bool
    "bool : Wether the source should loop / repeat or not. Defaults to false."
    LoopCount: int
    "int : How many times the source should loop / repeat.  Defaults to 0."
    Name: str
    "str : The name of the source."
    Type: ReaderType or str
    "ReaderType : The type of the source."
    Args: RequestArgs
    "RequestArgs : The additional arguments for the source."


class SourceData(TypedDict):
    """A source's raw data, as a payload."""

    Name: str
    "str : The name of the source."
    Type: str
    "str : The type of the source."
    Volume: float
    "float : The volume of the source."
    Duration: int
    "int : The duration of the source in milliseconds."
    Remaining: int
    "int : How many more milliseconds the source will be playing."
    Paused: bool
    "bool : Wether the source is paused or not."
    Loop: int
    "int : How many times the source will loop. If 0, the source will not repeat itself."
    ID: int
    "int : The ID of the source."
    EndTime: str
    "str : The estimated timestamp for when the source will finish playing."
    StartTime: str
    "str : When the source started playing."
    Request: RequestData
    "RequestData : The request used to create the source."


class AudioStatus(TypedDict):
    "The raw data read from /tmp/audioStatus.json."
    Sources: List[SourceData] or None
    "List[SourceData] : The sources that are know to the audio manager."
    Running: bool
    "bool : Wether the audio manager knows any sources or not."
    Disabled: bool
    "bool : Wether the audio manager is disabled or not."



"""Make apps quickly using Python."""

import os

import flask
from werkzeug.local import LocalProxy

from . import files
from . import html
from .app import ReplitApp
from .files import File
from .html import HTMLElement, Link, Page, Paragraph
from .utils import (
    authed_ratelimit,
    chain_decorators,
    find,
    local_redirect,
    needs_params,
    needs_sign_in,
    sign_in,
    sign_in_page,
    sign_in_snippet,
)
from ..database import AsyncDatabase, AsyncJSONKey, Database, db, JSONKey

auth = LocalProxy(lambda: flask.request.auth)
signed_in = LocalProxy(lambda: flask.request.signed_in)
request = LocalProxy(lambda: flask.request)
render_template = flask.render_template
redirect = flask.redirect


def user_data(username: str) -> JSONKey:
    """Shorthand for db.jsonkey(username, dict).

    Args:
        username (str): The key to use for the JSONKey.

    Returns:
        JSONKey: An initialized JSONKey.
    """
    return db.jsonkey(username, dict)


current_user_data = LocalProxy(lambda: user_data(flask.request.auth.name))



# flake8: noqa
from typing import Any, Callable, Dict, Optional, Tuple, Union
import json

JSON_TYPE = Optional[Union[str, int, float, bool, dict, list]]


class AsyncJSONKey:
    """Represents an key in the async database that holds a JSON value.
    db.jsonkey() will initialize an instance for you,
    you don't have to do it manually.
    """

    __slots__ = ("db", "key", "dtype", "get_default", "discard_bad_data", "do_raise")

    def __init__(
        self,
        db: Any,
        key: str,
        dtype: JSON_TYPE,
        get_default: Callable[[], JSON_TYPE] = None,
        discard_bad_data: bool = False,
        do_raise: bool = False,
    ) -> None:
        """Initialize the key.
        Args:
            db (Any): An instance of ReplitDb
            key (str): The key to read
            dtype (JSON_TYPE): The datatype the key should be, can be typing.Any.
            get_default (Callable): A function that returns the default
                value if the key is not set. If it is None (the default) the dtype
                argument is used.
            discard_bad_data (bool): Don't prompt if bad data is read, overwrite it
                with the default. Defaults to False.
            do_raise (bool): Whether to raise exceptions when errors are encountered.
        """
        self.db = db
        self.key = key
        self.dtype = dtype
        self.get_default = get_default
        self.discard_bad_data = discard_bad_data
        self.do_raise = do_raise

    def _default(self) -> JSON_TYPE:
        if self.get_default:
            return self.get_default()
        return self.dtype

    def _is_valid_type(self, data: JSON_TYPE) -> bool:
        return self.dtype is Any or isinstance(data, self.dtype)

    def _type_mismatch_msg(self, data: Any) -> str:
        return (
            f"Type mismatch: Got type {type(data).__name__},"
            f"expected {self.dtype.__name__}"
        )

    async def get(self) -> JSON_TYPE:
        """Get the value of the key.
        If an invalid JSON value is read or the type does not match, it will show a
            prompt asking the user what to do unless discard_bad_data is set.
        Raises:
            KeyError: If do_raise is true and the key does not exist.
            json.JSONDecodeError: If do_raise is true and invalid JSON data is read
                from the key.
        Returns:
            JSON_TYPE: The value read from the database
        """
        try:
            read = await self.db.get(self.key)
        except KeyError:
            if self.do_raise:
                raise
            print(f"Database key {self.key} not set, setting it to default value")
            default = self._default()
            await self.db.set(self.key, default)
            return default

        try:
            data = json.loads(read)
        except json.JSONDecodeError:
            if self.do_raise:
                raise
            return await self._error("Invalid JSON data read", read)

        if not self._is_valid_type(data):
            return await self._error(
                self._type_mismatch_msg(data),
                read,
            )
        return data

    async def _error(self, error: str, read: str) -> JSON_TYPE:
        if self.do_raise:
            raise ValueError(error)

        print(f"Error reading key {self.key!r}: {error}", file=stderr)
        if self.discard_bad_data:
            val = self._default()
            await self.db.set(self.key, json.dumps(val))
            print(f"Wrote default to key {self.key!r}")
            return val
        return await self._should_discard_prompt(error, read)

    async def _should_discard_prompt(self, error: str, read: str) -> bool:
        while True:
            choice = input(
                "d to use default, v to view the invalid data, c to insert custom "
                "value, ^C to exit: "
            )
            if choice.startswith("d"):
                print("Writing default...")
                val = self._default()
                await self.db.set(self.key, val)
                return val
            elif choice.startswith("v"):
                print(f"Data read from key: {read!r}")
            elif choice.startswith("c"):
                toset = input(
                    f"Enter data to write, should be of type {self.dtype.__name__!r}"
                    " (leave blank to return to menu): "
                )
                if not toset:
                    continue
                try:
                    data = json.loads(toset)
                except json.JSONDecodeError:
                    print("Invalid JSON data!")
                else:
                    if not self._is_valid_type(data):
                        print(self._type_mismatch_msg(data))
                        continue

                    await self.db.set(self.key, toset)
                    print("Wrote data to key")
                    return data

    async def set(self, data: JSON_TYPE) -> None:
        """Set the value of the jsonkey.
        Args:
            data (JSON_TYPE): The value to set it to.
        Raises:
            TypeError: The type of the value set does not match the datatype.
        """
        if not self._is_valid_type(data):
            raise TypeError(self._type_mismatch_msg(data))

        await self.db.set(self.key, json.dumps(data))


class JSONKey(AsyncJSONKey):
    """Represents a key in the database that holds a JSON value.

    db.jsonkey() will initialize an instance for you,
    you don't have to do it manually.
    """

    __slots__ = ("db", "key", "dtype", "get_default", "discard_bad_data", "do_raise")

    def __init__(
        self,
        db: Any,
        key: str,
        dtype: JSON_TYPE,
        get_default: Callable = None,
        discard_bad_data: bool = False,
        do_raise: bool = False,
    ) -> None:
        """Initialize the key.

        Args:
            db (Any): An instance of ReplitDb
            key (str): The key to read
            dtype (JSON_TYPE): The datatype the key should be, can be typing.Any.
            get_default (Callable): A function that returns the default
                value if the key is not set. If it is None (the default) the dtype
                argument is used.
            discard_bad_data (bool): Don't prompt if bad data is read, overwrite it
                with the default. Defaults to False.
            do_raise (bool): Whether to raise exceptions when errors are encountered.
        """
        self.db = db
        self.key = key
        self.dtype = dtype
        self.get_default = get_default
        self.discard_bad_data = discard_bad_data
        self.do_raise = do_raise

    def get(self) -> JSON_TYPE:
        """Get the value of the key.

        If an invalid JSON value is read or the type does not match, it will show a
            prompt asking the user what to do unless discard_bad_data is set.

        Returns:
            JSON_TYPE: The value read from the database
        """
        try:
            read = self.db[self.key]
        except KeyError:
            print(f"Database key {self.key} not set, setting it to default value")
            default = self._default()
            self.db[self.key] = default
            return default

        if isinstance(self.db, ReplitDb):
            try:
                data = json.loads(read)
            except json.JSONDecodeError:
                return self._error("Invalid JSON data read", read)
        else:
            data = read

        if not self._is_valid_type(data):
            return self._error(
                self._type_mismatch_msg(data),
                read,
            )
        return data

    def _error(self, error: str, read: str) -> JSON_TYPE:
        print(f"Error reading key {self.key!r}: {error}", file=stderr)
        if self.discard_bad_data:
            val = self._default()
            self.db[self.key] = json.dumps(val)
            print(f"Wrote default to key {self.key!r}")
            return val
        return self._should_discard_prompt(error, read)

    def _should_discard_prompt(self, error: str, read: str) -> bool:
        while True:
            choice = input(
                "d to use default, v to view the invalid data, c to insert custom "
                "value, ^C to exit: "
            )
            if choice.startswith("d"):
                print("Writing default...")
                val = self._default()
                self.db[self.key] = val
                return val
            elif choice.startswith("v"):
                print(f"Data read from key: {read!r}")
            elif choice.startswith("c"):
                toset = input(
                    f"Enter data to write, should be of type {self.dtype.__name__!r}"
                    " (leave blank to return to menu): "
                )
                if not toset:
                    continue
                try:
                    data = json.loads(toset)
                except json.JSONDecodeError:
                    print("Invalid JSON data!")
                else:
                    if not self._is_valid_type(data):
                        print(self._type_mismatch_msg(data))
                        continue

                    self.db[self.key] = toset
                    print("Wrote data to key")
                    return data

    def set(self, data: JSON_TYPE) -> None:
        """Set the value of the jsonkey.

        Args:
            data (JSON_TYPE): The value to set it to.

        Raises:
            TypeError: The type of the value set does not match the datatype.
        """
        if not self._is_valid_type(data):
            raise TypeError(self._type_mismatch_msg(data))
        if isinstance(self.db, ReplitDb):
            data = json.dumps(data)
        self.db[self.key] = data

    def read(self, key: str, default: Any = None) -> Any:
        """Shorthand for self.get().get(name, default) if datatype is dict.

        Args:
            key (str): The name to get.
            default (Any): The default if the key doesn't exist. Defaults to None.

        Returns:
            Any: The value read or the default.
        """
        data = self.get()
        if not isinstance(data, dict):
            raise TypeError
        return data.get(key, default)

    def keys(self, *keys: str) -> Any:
        """Reads multiple keys from the key's value and allows setting.

        Args:
            *keys (str): The keys to read from the data.

        Returns:
            Any: The value accessed from self.get()[k1][k2][kn]
        """
        data = self
        for key in keys[:-1]:
            data = type(self)(db=data, key=key, dtype=Any)
        check = data[keys[-1]]
        if type(check) is dict:
            return type(self)(db=data, key=keys[-1], dtype=dict)
        else:
            return check

    def __getitem__(self, name: str) -> JSON_TYPE:
        """Retrieve a key from the JSONKey's value if it is a dict.

        Args:
            name (str): The name to retrieve.

        Returns:
            JSON_TYPE: The value of the key.
        """
        return self.keys(name)

    def __setitem__(self, name: str, value: JSON_TYPE) -> None:
        """Sets a key inside the JSONKey's value if it is a dict.

        Args:
            name (str): The key to set.
            value (JSON_TYPE): The value to set it to.
        """
        data = self.get()
        if not isinstance(data, dict):
            raise TypeError
        data[name] = value
        self.set(data)

    def append(self, item: JSON_TYPE) -> None:
        """Append to the JSONKey's value if it is a list.

        Args:
            item (JSON_TYPE): The item to append.
        """
        data = self.get()
        if not isinstance(data, list):
            raise TypeError
        self.set(data + [item])

    def __add__(self, item: Any) -> Any:
        """Add to the JSONKey's value and return the result.

        Args:
            item (Any): The item to add.

        Returns:
            Any: The result of adding.
        """
        return self.get() + item

    def __iadd__(self, item: Any) -> Any:
        """Add to the JSONKey's value and set the result.

        Args:
            item (Any): The item to add.

        Returns:
            Any: self
        """
        r = self.get() + item
        self.set(r)
        return self



"""Pure Python ANSI Color Escape Code generator."""

import colorsys


def clear() -> None:
    """Clear the terminal."""

    print("\033[H\033[2J", end="", flush=True)


class Color:
    """Dynamic Color: Accepts RGB Color."""

    def __init__(self, r: int, g: int, b: int) -> None:

        """
        Generate an ANSI escape code for color

        Args:
          r (int): Amount of red in color
          g (int): Amount of green in color
          b (int): Amount of blue in color

        Raises:
          ValueError
        """

        if r < 0 or g < 0 or b < 0:
            raise ValueError("No Color Support for colors under 0")
        if r > 255 or g > 255 or b > 255:
            raise ValueError("No Color Support for colors over 255")

        self.rgb = (r, g, b)
        self.fg = f"\033[38;2;{r};{g};{b}m"
        self.bg = f"\033[48;2;{r};{g};{b}m"

    @classmethod
    def hexdec(cls, hexvalue: str) -> None:
        """
        Convert Hex Value to RGB
        then generate an ANSI escape code

        Args:
          hexvalue (str): The color's hex value

        Raises:
          ValueError

        Returns:
          color : RGB colors from Hex Value
        """

        try:
            hexvalue = hexvalue.lstrip("#")
            r, g, b = tuple(int(hexvalue[i : i + 2], 16) for i in (0, 2, 4))
        except Exception as e:
            raise ValueError(f"Error while converting Hex to RGB - {e}")

        return cls(r, g, b)

    @classmethod
    def hsv(cls, h, s, v) -> None:
        """
        Convert Hex Value to RGB
        then generate an ANSI escape code

        Args:
          hexvalue (str): The color's hex value

        Raises:
          ValueError

        Returns:
          color : RGB colors from Hex Value
        """
        try:
            r, g, b = colorsys.hsv_to_rgb(h, s, v)
        except:
            raise ValueError("Converting HSV to RGB ran into an error")

        return cls(r, g, b)

    @classmethod
    def hls(cls, h, l, s):
        """
        Convert Hex Value to RGB
        then generate an ANSI escape code

        Args:
          hexvalue (str): The color's hex value

        Raises:
          ValueError

        Returns:
          color : RGB colors from Hex Value
        """
        try:
            r, g, b = colorsys.hls_to_rgb(h, s, v)
        except:
            raise ValueError("Converting HLS to RGB ran into an error")

        return cls(r, g, b)


class Bit:
    def __init__(self, value: int) -> None:
        """
        Use a 8bit colorpallete
        colors from  0-255 (256 total)

        Args:
          value (int): A color value from 0 to 255

        Raises:
          ValueError
        """

        if value > 255:
            raise ValueError("8 Bit Pallete - No Color Support for Colors over 255")
        if value < 0:
            raise ValueError("8 Bit Pallete - No Color Support for colors under 0")

        self.fg = f"\033[38;5;{value}m"
        self.bg = f"\033[48;5;{value}m"


attributes = (
    {  # use only repl.it supported ansi codes. Codes such as blink do not work.
        "reset": 0,
        "bold": 1,
        "faint": 2,
        "italic": 3,
        "underline": 4,
        "highlight": 7,
    }
)


class Attr:
    def __init__(self, attrib: str) -> None:
        """
        Custom Attributes such as bold and italic
        Returns ANSI Escape

        Args:
          attrib (str): Special Styles for characters

        Raises:
          ValueError
        """

        if attrib in attributes:
            self.attr = f"\033[{attributes[attrib]}m"
        else:
            raise ValueError(f"Attributes - {attrib} is not supported.")


reset = Attr("reset").attr
bold = Attr("bold").attr
italic = Attr("italic").attr
red = Color(255, 0, 0)
orange = Color(255, 165, 0)
yellow = Color(255, 255, 0)
green = Color(0, 255, 0)
blue = Color(0, 0, 255)
indigo = Color(75, 0, 130)
violet = Color(238, 130, 238)
purple = Color(128, 0, 128)
pink = Color(255, 105, 180)
brown = Color(165, 42, 42)
brightred = Color(250, 128, 114)
brightorange = Color(255, 215, 0)
brightyellow = Color(255, 255, 102)
brightgreen = Color(102, 255, 102)
brightblue = Color(102, 178, 255)
brightpurple = Color(178, 102, 255)
darkred = Color(139, 0, 0)
darkorange = Color(255, 140, 0)
darkyellow = Color(204, 204, 0)
darkgreen = Color(0, 153, 0)
darkblue = Color(0, 0, 204)
darkpurple = Color(102, 0, 204)



"""A library to play audio in a repl."""
from datetime import datetime, timedelta
import json
from os import path
import time
from typing import Any, List, Optional

from .types import (
    AudioStatus,
    file_types,
    ReaderType,
    RequestArgs,
    RequestData,
    SourceData,
    WaveType,
)


class InvalidFileType(Exception):
    """Exception for when a requested file's type isnt valid."""

    pass


class NoSuchSourceException(Exception):
    """Exception used when a source doesn't exist."""

    pass


class Source:
    """A Source is used to get audio that is sent to the user."""

    __payload: SourceData
    _loops: bool
    _name: str

    def __init__(self, payload: SourceData, loops: bool) -> None:
        """Initialize the class.

        Args:
            payload (SourceData): The payload for the source.
            loops (bool): How many times the source should loop.
        """
        self.__payload = payload
        self._loops = loops
        self._name = payload["Name"]

    def __get_source(self) -> SourceData or None:
        source = None
        with open("/tmp/audioStatus.json", "r") as f:
            data = json.loads(f.read())
            for s in data["Sources"]:
                if s["ID"] == self.id:
                    source = s
                    break
        if source:
            self.__payload = source
        return source

    def __update_source(self, **changes: Any) -> None:
        s = self.__get_source()
        if not s:
            raise NoSuchSourceException(
                f'No player with id "{id}" found! It might be done playing.'
            )
        s.update({key.title(): changes[key] for key in changes})
        with open("/tmp/audio", "w") as f:
            f.write(json.dumps(s))
        self.__get_source()

    @property
    def name(self) -> str:
        """The name of the source."""
        return self._name

    def get_start_time(self) -> datetime:
        """When the source started plaing."""
        timestamp_str = self.__payload["StartTime"]
        timestamp = datetime.strptime(timestamp_str[:-4], "%Y-%m-%dT%H:%M:%S.%f")
        return timestamp

    start_time: datetime = property(get_start_time)
    "Property wrapper for :py:meth:`~replit.Source.get_start_time`"

    @property
    def path(self) -> str or None:
        """The path to the source, if available."""
        data = self.__payload
        if ReaderType(data["Type"]) in file_types:
            return self.__payload["Request"]["Args"]["Path"]

    @property
    def id(self) -> int:
        """The ID of the source."""
        return self.__payload["ID"]

    def get_remaining(self) -> timedelta:
        """The estimated time remaining in the source's current loop."""
        data = self.__get_source()
        if not data:
            return timedelta(milliseconds=0)

        return timedelta(milliseconds=data["Remaining"])

    remaining: int = property(get_remaining)
    "Property wrapper for :py:meth:`~replit.Source.get_remaining`"

    def get_end_time(self) -> Optional[datetime]:
        """The estimated time when the source will be done playing.

        Returns:
            Optional[datetime]: The estimated time when the source will be done playing
                or None if it is already finished.
        """
        s = self.__get_source()
        if not s:
            return None

        timestamp_str = s["EndTime"]
        timestamp = datetime.strptime(timestamp_str[:-4], "%Y-%m-%dT%H:%M:%S.%f")
        return timestamp

    end_time: datetime or None = property(get_end_time)
    "Property wrapper for :py:meth:`~replit.Source.get_end_time`"

    @property
    def does_loop(self) -> bool:
        """Whether the source repeats itself or not."""
        return self._loops

    @property
    def duration(self) -> timedelta:
        """The duration of the source."""
        return timedelta(millaseconds=self.__payload["Duration"])

    def get_volume(self) -> float:
        """The volume the source is set to."""
        self.__get_source()
        return self.__payload["Volume"]

    def set_volume(self, volume: float) -> None:
        """Set the volume.

        Args:
            volume (float): The volume the source should be set to.
        """
        self.__update_source(volume=volume)

    volume: float = property(get_volume, set_volume)
    "Property wrapper for `replit.Source.get_volume` and `replit.Source.set_volume`"

    def get_paused(self) -> bool:
        """Whether the source is paused."""
        self.__get_source()
        return self.__payload["Paused"]

    def set_paused(self, paused: bool) -> None:
        """Change if the source is paused.

        Args:
            paused (bool): Whether the source should be paused.
        """
        self.__update_source(paused=paused)

    paused = property(get_paused, set_paused)
    "Property wrapper for `replit.Source.get_paused` and `replit.Source.set_paused`"

    def get_loops_remaining(self) -> Optional[int]:
        """The remaining amount of times the file will restart.

        Returns:
            Optional[int]: The number of loops remaining or None if the source can't be
                found, either because it has finished playing or an error occured.
        """
        if not self._loops:
            return 0

        s = self.__get_source()
        if not s:
            return None

        if s["ID"] == self.id:
            loops = s["Loop"]

            return loops

    def set_loop(self, loop_count: int) -> None:
        """Set the remaining amount of loops for the source.

        Args:
            loop_count (int): How many times the source should repeat itself. Set to a
                negative value for infinite.
        """
        does_loop = loop_count != 0
        self._loops = does_loop
        self.__update_source(doesLoop=does_loop, loopCount=loop_count)

    loops_remaining: int or None = property(get_loops_remaining)
    "Property wrapper for :py:meth:`~replit.Source.get_loops_remaining`"

    def toggle_playing(self) -> None:
        """Play/pause the source."""
        self.set_paused(not self.paused)


class Audio:
    """The basic audio manager.

    Notes
    -----
    This is not intended to be called directly, instead use :py:const:`audio`.

    Using this in addition to `audio` can cause **major** issues.
    """

    __known_ids = []
    __names_created = 0

    def __gen_name(self) -> str:
        return f"Source {time.time()}"

    def __get_new_source(self, name: str, does_loop: bool) -> Source:
        new_source = None
        timeOut = datetime.now() + timedelta(seconds=2)

        while not new_source and datetime.now() < timeOut:
            try:
                sources = AudioStatus(self.read_status())["Sources"]
                new_source = SourceData([s for s in sources if s["Name"] == name][0])
            except IndexError:
                pass
            except json.JSONDecodeError:
                pass

        if not new_source:
            raise TimeoutError("Source was not created within 2 seconds.")

        return Source(new_source, does_loop)

    def play_file(
        self,
        file_path: str,
        volume: float = 1,
        does_loop: bool = False,
        loop_count: int = 0,
        name: Optional[str] = None,
    ) -> Source:
        """Sends a request to play a file, assuming the file is valid.

        Args:
            file_path (str): The path to the file that should be played. Can be
                absolute or relative.
            volume (float): The volume the source should be played at. (1 being
                100%)
            does_loop (bool): Wether the source should repeat itself or not. Note, if
                you set this you should also set loop_count.
            loop_count (int): How many times the source should repeat itself. Set to 0
                to have the source play only once, or set to a negative value for the
                source to repeat forever.
            name (str): The name of the file. Default value is a unique name for the
                source.

        Returns:
            Source: The source created with the provided data.

        Raises:
            FileNotFoundError: If the file is not found.
            InvalidFileType: If the file type is not valid.
        """
        name = name or self.__gen_name()

        if not path.exists(file_path):
            raise FileNotFoundError(f'File "{file_path}" not found.')

        file_type = file_path.split(".")[-1]

        if ReaderType(file_type) not in file_types:
            raise InvalidFileType(f"Type {file_type} is not supported.")

        data = RequestData(
            Type=file_type,
            Volume=volume,
            DoesLoop=does_loop,
            LoopCount=loop_count,
            Name=name,
            Args=RequestArgs(Path=file_path),
        )

        with open("/tmp/audio", "w") as p:
            p.write(json.dumps(dict(data)))

        return self.__get_new_source(name, does_loop)

    def play_tone(
        self,
        duration: float,
        pitch: int,
        wave_type: WaveType,
        does_loop: bool = False,
        loop_count: int = 0,
        volume: float = 1,
        name: Optional[str] = None,
    ) -> Source:
        """Play a tone from a frequency and wave type.

        Args:
            duration (float): How long the tone should be played (in seconds).
            pitch (int): The frequency the tone should be played at.
            wave_type (WaveType): The wave shape used to generate the tone.
            does_loop (bool): Wether the source should repeat itself or not. Note, if
                you set this you should also set loop_count.
            loop_count (int): How many times the source should repeat itself. Set to 0
                to have the source play only once, or set to a negative value for the
                source to repeat forever.
            volume (float): The volume the tone should be played at (1 being 100%).
            name (str): The name of the file. Default value is a unique name for the
                source.

        Returns:
            Source: The source for the tone.
        """
        name = name or self.__gen_name()

        # ensure the wave type is valid. This will throw an error if it isn't.
        WaveType(wave_type)

        data = RequestData(
            Name=name,
            DoesLoop=does_loop,
            LoopCount=loop_count,
            Volume=volume,
            Type=str(ReaderType.tone),
            Args=RequestArgs(
                WaveType=wave_type,
                Pitch=pitch,
                Seconds=duration,
            ),
        )

        with open("/tmp/audio", "w") as f:
            f.write(json.dumps(data))

        return self.__get_new_source(name, does_loop)

    def get_source(self, source_id: int) -> Source or None:
        """Get a source by it's ID.

        Args:
            source_id (int): The ID for the source that should be found.

        Raises:
            NoSuchSourceException: If the source isnt found or there isn't any sources
                known to the audio manager.

        Returns:
            Source: The source with the ID provided.
        """
        source = None
        with open("/tmp/audioStatus.json", "r") as f:
            data = AudioStatus(json.loads(f.read()))
            if not data["Sources"]:
                raise NoSuchSourceException("No sources exist yet.")
            for s in data["Sources"]:

                if s["ID"] == int(source_id):
                    source = s
                    break
        if not source:
            raise NoSuchSourceException(f'Could not find source with ID "{source_id}"')
        return Source(source, source["Loop"])

    def read_status(self) -> AudioStatus:
        """Get the raw data for what's playing.

        This is an api call, and shouldn't be needed for general usage.

        Returns:
            AudioStatus: The contents of /tmp/audioStatus.json
        """
        with open("/tmp/audioStatus.json", "r") as f:
            data = AudioStatus(json.loads(f.read()))
            if data["Sources"] is None:
                data["Sources"]: List[SourceData] = []
            return data

    def get_playing(self) -> List[Source]:
        """Get a list of playing sources.

        Returns:
            List[Source]: A list of sources that aren't paused.
        """
        data = self.read_status()
        sources = data["Sources"]
        return [Source(s, s["Loop"]) for s in sources if not s["Paused"]]

    def get_paused(self) -> List[Source]:
        """Get a list of paused sources.

        Returns:
            List[Source]: A list of sources that are paused.
        """
        data = self.read_status()
        sources = data["Sources"]
        return [Source(s, s["Loop"]) for s in sources if s["Paused"]]

    def get_sources(self) -> List[Source]:
        """Gets all sources.

        Returns:
            List[Source]: Every source known to the audio manager, paused or playing.
        """
        data = self.read_status()
        sources = data["Sources"]
        return [Source(s, s["Loop"]) for s in sources]



"""Utilities for interacting with static files."""

import flask


cache = {}


class File(flask.Response):
    """Represents a static file."""

    def __init__(self, filename: str, no_cache: bool = False) -> None:
        """Initialize the file.

        Args:
            filename (str): The filename to read from.
            no_cache (bool): Whether to skip caching the file. Defaults to
                False.
        """
        self.filename = str(filename)
        self.no_cache = no_cache

        if filename in cache:
            self.content = cache[filename]
        else:
            with open(filename, "r") as f:
                self.content = f.read()

            if not no_cache:
                cache[filename] = self.content

        super().__init__(self.content)



# flake8: noqa
import time
import unittest
import replit
from .. import audio
from . import types

test_file = "../test.mp3"


class TestAudio(unittest.TestCase):
    def test_creation(self):
        source = audio.play_file(test_file)
        self.assertEqual(source.path, test_file)
        source.paused = True
        time.sleep(1)
        self.assertEqual(source.paused, True, "Pausing Source")

    def test_pause(self):
        source = audio.play_file(test_file)
        source.volume = 2
        time.sleep(1)
        self.assertEqual(source.volume, 2, "Volume set to 2")

        source.paused = True
        time.sleep(1)
        self.assertEqual(source.paused, True, "Pausing Source")

        source.volume = 0.2
        time.sleep(1)
        self.assertEqual(source.volume, 0.2, "Volume set to .2")

        source.paused = True
        time.sleep(1)
        self.assertEqual(source.paused, True, "Pausing Source")

    def test_loop_setting(self):
        source = audio.play_file(test_file)

        self.assertEqual(source.loops_remaining, 0, "0 loops remaining")
        source.set_loop(2)
        time.sleep(1)

        self.assertEqual(source.loops_remaining, 2, "2 loops remaining")
        source.paused = True
        time.sleep(1)
        self.assertEqual(source.paused, True, "Pausing Source")

    def test_other(self):
        source = audio.play_file(test_file)

        self.assertIsNotNone(source.end_time)
        self.assertIsNotNone(source.start_time)
        self.assertIsNotNone(source.remaining)
        source.paused = True
        time.sleep(1)
        self.assertEqual(source.paused, True, "Pausing Source")

    def test_tones(self):
        try:
            audio.play_tone(2, 400, 2)
        except TimeoutError or ValueError as e:
            self.fail(e)


if __name__ == "__main__":
    unittest.main()



"""Async and dict-like interfaces for interacting with Repl.it Database."""

from collections import abc
import json
from typing import Any, Dict, Iterator, Tuple
import urllib

import aiohttp
import requests


class AsyncDatabase:
    """Async interface for Repl.it Database."""

    __slots__ = ("db_url", "sess")

    def __init__(self, db_url: str) -> None:
        """Initialize database. You shouldn't have to do this manually.

        Args:
            db_url (str): Database url to use.
        """
        self.db_url = db_url

    async def get(self, key: str) -> str:
        """Get the value of an item from the database.

        Args:
            key (str): The key to retreive

        Raises:
            KeyError: Key is not set

        Returns:
            str: The value of the key
        """
        async with aiohttp.ClientSession() as session:
            async with session.get(
                self.db_url + "/" + urllib.parse.quote(key)
            ) as response:
                if response.status == 404:
                    raise KeyError(key)
                response.raise_for_status()
                return await response.text()

    async def set(self, key: str, value: str) -> None:
        """Set a key in the database to value.

        Args:
            key (str): The key to set
            value (str): The value to set it to
        """
        async with aiohttp.ClientSession() as session:
            async with session.post(self.db_url, data={key: value}) as response:
                response.raise_for_status()

    async def delete(self, key: str) -> None:
        """Delete a key from the database.

        Args:
            key (str): The key to delete
        """
        async with aiohttp.ClientSession() as session:
            async with session.delete(
                self.db_url + "/" + urllib.parse.quote(key)
            ) as response:
                response.raise_for_status()

    async def list(self, prefix: str) -> Tuple[str, ...]:
        """List keys in the database which start with prefix.

        Args:
            prefix (str): The prefix keys must start with, blank not not check.

        Returns:
            Tuple[str]: The keys found.
        """
        params = {"prefix": prefix, "encode": "true"}
        async with aiohttp.ClientSession() as session:
            async with session.get(self.db_url, params=params) as response:
                response.raise_for_status()
                text = await response.text()
                if not text:
                    return tuple()
                else:
                    return tuple(urllib.parse.unquote(k) for k in text.split("\n"))

    async def to_dict(self, prefix: str = "") -> Dict[str, str]:
        """Dump all data in the database into a dictionary.

        Args:
            prefix (str): The prefix the keys must start with,
                blank means anything. Defaults to "".

        Returns:
            Dict[str, str]: All keys in the database.
        """
        ret = {}
        keys = await self.list(prefix=prefix)
        for i in keys:
            ret[i] = await self.get(i)
        return ret

    async def keys(self) -> Tuple[str, ...]:
        """Get all keys in the database.

        Returns:
            Tuple[str]: The keys in the database.
        """
        return await self.list("")

    async def values(self) -> Tuple[str, ...]:
        """Get every value in the database.

        Returns:
            Tuple[str]: The values in the database.
        """
        data = await self.to_dict()
        return tuple(data.values())

    async def items(self) -> Tuple[Tuple[str, str], ...]:
        """Convert the database to a dict and return the dict's items method.

        Returns:
            Tuple[Tuple[str]]: The items
        """
        return tuple((await self.to_dict()).items())

    def __repr__(self) -> str:
        """A representation of the database.

        Returns:
            A string representation of the database object.
        """
        return f"<{self.__class__.__name__}(db_url={self.db_url!r})>"


class Database(abc.MutableMapping):
    """Dictionary-like interface for Repl.it Database.

    This interface will coerce all values everything to and from JSON. If you
    don't want this, use AsyncDatabase instead.
    """

    __slots__ = ("db_url", "sess")

    def __init__(self, db_url: str) -> None:
        """Initialize database. You shouldn't have to do this manually.

        Args:
            db_url (str): Database url to use.
        """
        self.db_url = db_url
        self.sess = requests.Session()

    def __getitem__(self, key: str) -> Any:
        """Get the value of an item from the database.

        Args:
            key (str): The key to retreive

        Raises:
            KeyError: Key is not set

        Returns:
            Any: The value of the key
        """
        r = self.sess.get(f"{self.db_url}/{key}")
        if r.status_code == 404:
            raise KeyError(key)

        r.raise_for_status()
        return json.loads(r.text)

    def __setitem__(self, key: str, value: Any) -> None:
        """Set a key in the database to value.

        Args:
            key (str): The key to set
            value (Any): The value to set it to. Must be JSON-serializable.
        """
        j = json.dumps(value, separators=(",", ":"))
        r = self.sess.post(self.db_url, data={key: j})
        r.raise_for_status()

    def __delitem__(self, key: str) -> None:
        """Delete a key from the database.

        Args:
            key (str): The key to delete

        Raises:
            KeyError: Key is not set
        """
        r = self.sess.delete(f"{self.db_url}/{key}")
        if r.status_code == 404:
            raise KeyError(key)

        r.raise_for_status()

    def __iter__(self) -> Iterator[str]:
        """Return an iterator for the database."""
        return iter(self.prefix(""))

    def __len__(self) -> int:
        """The number of keys in the database."""
        return len(self.prefix(""))

    def prefix(self, prefix: str) -> Tuple[str, ...]:
        """Return all of the keys in the database that begin with the prefix.

        Args:
            prefix (str): The prefix the keys must start with,
                blank means anything.

        Returns:
            Tuple[str]: The keys found.
        """
        r = requests.get(f"{self.db_url}", params={"prefix": prefix, "encode": "true"})
        r.raise_for_status()

        if not r.text:
            return tuple()
        else:
            return tuple(urllib.parse.unquote(k) for k in r.text.split("\n"))

    def __repr__(self) -> str:
        """A representation of the database.

        Returns:
            A string representation of the database object.
        """
        return f"<{self.__class__.__name__}(db_url={self.db_url!r})>"



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['replitdev', 'replitdev.audio', 'replitdev.database', 'replitdev.web']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.6.2,<4.0.0',
 'flask>=1.1.2,<2.0.0',
 'requests-html>=0.10.0,<0.11.0',
 'typing_extensions>=3.7.4,<4.0.0',
 'werkzeug>=1.0.1,<2.0.0']

setup_kwargs = {
    'name': 'replitdev',
    'version': '2.2.0',
    'description': 'A library for interacting with features of repl.it',
    'long_description': '# `replit` Python Module\n\n\nThe [Repl.it](https://repl.it/) Python environment does not require any platform-specific code, however, these optional utilities provide additional platform features in a simple and accessible way.\n\n![compute](https://github.com/kennethreitz42/replit-py/blob/kr-cleanup/ext/readme.gif?raw=true)\n\n\n*Example*: [Repl.it DB](https://docs.repl.it/misc/database) is an HTTP service, but an optional Python client (here!) is available.\n\n\n### `>>> import replit`\n\nThis repository is the home for the `replit` Python package, which provides:\n\n- A fully-featured database client for [Repl.it DB](https://docs.repl.it/misc/database). **[[docs]](https://example.com)**\n- A **work in progress** Repl.it user profile lookup. **[[docs]](https://example.com)**\n- A Flask application decorator for ensuring Repl.it Auth required on specific routes. **[[docs]](https://example.com)**\n\n& other helpful toys and utilities, like...\n\n- A simple audio library that can play tones and audio files!\n- Some helpful functions for displaying ANSI colors within interactive terminal sessions.\n\n### Open Source License\n\nThis library is licensed under the [ISC License](https://en.wikipedia.org/wiki/ISC_license) and is free for you to use, change, or even profit from!\n',
    'author': 'mat',
    'author_email': 'pypi@matdoes.dev',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/replit/replit-py',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from requests import Session as HTTPSession
from requests_html import HTMLSession

http = HTTPSession()
html_http = HTMLSession()

# TODO: provide a way to list repls.


class ReplitUser:
    def __init__(self):
        self.username = None
        self.name = None
        self.bio = None
        self.avatar_url = None
        self.languages = None

    @classmethod
    def from_username(_class, username):
        """Creates a new ReplitUser object from a given Repl.it profile name."""
        # TODO: catch non-existient users.
        url = _class._replit_url_from_username(username)
        return _class.from_replit_profile_url(url)

    @classmethod
    def from_replit_profile_url(_class, url):
        """Creates a new ReplitUser object from a given Repl.it profile URL."""

        # Fetch the profile from the web, and encapsulate its HTML.
        r = html_http.get(url=url)
        html = r.html

        # Instantiate the class.
        user = _class()

        # Populate the user instance from parsed HTML.
        user.name = user.__extract_name(html)
        user.bio = user.__extract_bio(html)
        user.avatar_url = user.__extract_avatar_url(html)

        return user

    @property
    def avatar_content(self):
        if self.avatar_url:
            return http.get(self.avatar_url).content

    @staticmethod
    def _replit_url_from_username(username):
        return f"https://repl.it/@{username}"

    def __extract_name(self, html):
        return html.find("h1", first=True).text

    def __extract_bio(self, html):
        return html.find(".Linkify", first=True).text

    def __extract_avatar_url(self, html):
        e = html.find(".profile-icon > span", first=True)
        style_str = e.attrs["style"]

        # Remove the CSS-y parts of the image URL, in a clear manner.
        len_0 = len('background-image:url("')
        len_1 = len('"]')

        # Slice the background string up.
        avatar_url = style_str[len_0 : (len(style_str) - len_1)]

        return avatar_url


def get_user(username):
    """Creates a new ReplitUser object from a given Repl.it profile name."""

    return ReplitUser.from_username(username)

# Syntax suagar.
User = ReplitUser
